/*
 * Copyright (c) 1998-2001 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 *
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * Copyright (c) 2000-2001 Apple Computer, Inc.  All rights reserved.
 *
 *  DRI: Josh de Cesare
 *
 */

#include <IOKit/IOLib.h>
#include <vm/vm_kern.h>

#include "AppleBt8xx.h"
#include "AppleBt8xxRegs.h"

extern pmap_t kernel_pmap;

static OSDictionary *gAppleBt8xxInstances;
static IOLock *gAppleBt8xxInstancesLock;


//---------------------------------------------------------------------------//

#define super IOService
OSDefineMetaClassAndStructorsWithInit(AppleBt8xx, IOService, AppleBt8xx::metaInit());

void AppleBt8xx::metaInit(void)
{
  gAppleBt8xxInstances = OSDictionary::withCapacity(1);
  gAppleBt8xxInstancesLock = IOLockAlloc();
}

bool AppleBt8xx::init(OSDictionary *dictionary)
{
  bool res;
  
  res = super::init(dictionary);
  
  return res;
}

void AppleBt8xx::free(void)
{
  if (uniqueSymbol) uniqueSymbol->release();
  
  super::free();
}


bool AppleBt8xx::attach(IOService *provider)
{
  bool res;
  
  res = super::attach(provider);
  
  return res;
}

void AppleBt8xx::detach(IOService *provider)
{
  super::detach(provider);
}


IOService *AppleBt8xx::probe(IOService *provider, SInt32 *score)
{
  IOService      *instance, *tmpInstance;
  IOPCIDevice    *pciNub = (IOPCIDevice *)provider;
  long           uniqueNumber;
  char           uniqueName[9];
  OSObject       *tmpObject;
  
  instance = super::probe(provider, score);
  if (instance == 0) return 0;
  
  // Construct the unique name for this card.
  uniqueNumber = (pciNub->getBusNumber() << 16) |
                 (pciNub->getDeviceNumber() << 11);
  sprintf(uniqueName, "%08lx", uniqueNumber);
  uniqueSymbol = OSSymbol::withCString(uniqueName);
  
  IOLockLock(gAppleBt8xxInstancesLock);
  
  // Find the instance for this card.
  tmpObject = gAppleBt8xxInstances->getObject(uniqueSymbol);
  tmpInstance = OSDynamicCast(IOService, tmpObject);
  if (tmpInstance == 0) {
    gAppleBt8xxInstances->setObject(uniqueSymbol, instance);
  } else {
    instance = tmpInstance;
  }
  
  IOLockUnlock(gAppleBt8xxInstancesLock);
  
  return instance;
}

bool AppleBt8xx::start(IOService *provider)
{
  IOPCIDevice      *pciNub = (IOPCIDevice *)provider;
  IOMemoryMap      *pciMemoryMap;
  IOVirtualAddress pciBaseAddress;
  OSData           *tmpData;
  
  if (!super::start(provider)) return false;
  
  pciNub->setMemoryEnable(true);
  pciNub->setBusMasterEnable(true);
  pciMemoryMap = pciNub->mapDeviceMemoryWithRegister(kIOPCIConfigBaseAddress0);
  pciBaseAddress = pciMemoryMap->getVirtualAddress();
  
  _identifyChipType(pciNub);
  
  if (pciNub->getFunctionNumber() == 0) {
    videoNub = pciNub;
    videoMemoryMap = pciMemoryMap;
    videoBaseAddress = (UInt8 *)pciBaseAddress;
  } else {
    audioNub = pciNub;
    audioMemoryMap = pciMemoryMap;
    audioBaseAddress = (UInt8 *)pciBaseAddress;
  }
  
  // Wait for both nubs before starting on Bt878 and Bt879.
  if ((chipType == kBt878) || (chipType == kBt879)) {
    if ((videoNub == 0) || (audioNub == 0)) return true;
  }
  
  kmem_alloc_wired(kernel_map, (vm_offset_t *)&_videoCmdPage, PAGE_SIZE);
  kmem_alloc_wired(kernel_map, (vm_offset_t *)&_videoEvenPage, PAGE_SIZE);
  kmem_alloc_wired(kernel_map, (vm_offset_t *)&_videoOddPage, PAGE_SIZE);
  
  _videoCmdPagePhys  = (IOPhysicalAddress)pmap_extract(kernel_pmap, (vm_offset_t)_videoCmdPage);
  _videoEvenPagePhys = (IOPhysicalAddress)pmap_extract(kernel_pmap, (vm_offset_t)_videoEvenPage);
  _videoOddPagePhys  = (IOPhysicalAddress)pmap_extract(kernel_pmap, (vm_offset_t)_videoOddPage);
  
  _sources = OSDynamicCast(OSArray, getProperty("AppleBt8xxSources"));
  if (_sources == 0) return false;
  
  _audioMask = 0;
  tmpData = OSDynamicCast(OSData, getProperty("AppleBt8xxAudioMask"));
  if (tmpData != 0) _audioMask = *(UInt32 *)tmpData->getBytesNoCopy();
  
  _reset();
  
  registerService();
  
  return true;
}

void AppleBt8xx::stop(IOService *provider)
{
  IOPCIDevice *pciNub = (IOPCIDevice *)provider;
  IOMemoryMap *pciMemoryMap;
  
  _stop();
  
  if (pciNub->getFunctionNumber() == 0) {
    pciMemoryMap = videoMemoryMap;
    videoNub = 0;
  } else {
    pciMemoryMap = audioMemoryMap;
    audioNub = 0;
  }
  
  pciNub->setBusMasterEnable(false);
  pciNub->setMemoryEnable(false);
  pciMemoryMap->release();
  
  kmem_free(kernel_map, (vm_offset_t)_videoCmdPage, PAGE_SIZE);
  kmem_free(kernel_map, (vm_offset_t)_videoEvenPage, PAGE_SIZE);
  kmem_free(kernel_map, (vm_offset_t)_videoOddPage, PAGE_SIZE);
  
  // Wait for both nubs before removing the instance.
  if ((videoNub == 0) && (audioNub == 0)) {
    IOLockLock(gAppleBt8xxInstancesLock);
    gAppleBt8xxInstances->removeObject(uniqueSymbol);
    IOLockUnlock(gAppleBt8xxInstancesLock);
  }
  
  super::stop(provider);
}

IOReturn AppleBt8xx::setProperties(OSObject *properties)
{
  IOReturn           ret;
  OSData             *tmpData;
  AppleBt8xxCmdPtr   cmd;
  
  tmpData = OSDynamicCast(OSData, properties);
  cmd = (AppleBt8xxCmdPtr)tmpData->getBytesNoCopy();
  
  ret = _processCmd(cmd);
  
  return ret;
}

IOReturn AppleBt8xx::newUserClient(task_t owningTask, void *security_id,
				   UInt32 type, IOUserClient **handler)
{
  IOUserClient *newUserClient;
  
  if (_userClient != 0) return kIOReturnExclusiveAccess;
  
  newUserClient = AppleBt8xxUserClient::withTask(owningTask);
  if (newUserClient) {
    if (!newUserClient->attach(this) ||
	!newUserClient->start(this)) {
      newUserClient->detach(this);
      newUserClient->release();
    } else {
      _userClient = newUserClient;
    }
  }
  
  if (_userClient == 0) return kIOReturnNotOpen;
  
  *handler = _userClient;
  
  return kIOReturnSuccess;
}

void AppleBt8xx::close(void)
{
  _userClient = 0;
}


//---------------------------------------------------------------------------//

#define super IOService
OSDefineMetaClassAndStructors(AppleBt8xxUserClient, IOUserClient);

AppleBt8xxUserClient *AppleBt8xxUserClient::withTask(task_t owningTask)
{
  AppleBt8xxUserClient *userClient;
  
  userClient = new AppleBt8xxUserClient;;
  if(userClient && !userClient->init()) {
    userClient->release();
    userClient = 0;
  }
  
  return userClient;
}


#define SetExternalMethod(index, obj, methodName, type, countIn, countOut) \
externals[(index)].object = (obj); \
externals[(index)].func   = (IOMethod)&AppleBt8xxUserClient::##methodName; \
externals[(index)].flags  = (type); \
externals[(index)].count0 = (countIn); \
externals[(index)].count1 = (countOut);

bool AppleBt8xxUserClient::start(IOService *provider)
{
  if (!super::start(provider))
    return false;
  
  // Setup the external methods.
  SetExternalMethod(kAppleBt8xxCmdNone, this, _extCmdNone,
		    kIOUCScalarIScalarO, 0, 0);
  
  SetExternalMethod(kAppleBt8xxCmdSetup, this, _extCmdSetup,
		    kIOUCStructIStructO,
		    sizeof(AppleBt8xxCmdSetupInfo), 0);
  
  SetExternalMethod(kAppleBt8xxCmdStart, this, _extCmdStart,
		    kIOUCScalarIScalarO, 0, 0);
  SetExternalMethod(kAppleBt8xxCmdStop, this, _extCmdStop,
		    kIOUCScalarIScalarO, 0, 0);
  
  SetExternalMethod(kAppleBt8xxCmdGetSource, this, _extCmdGetSource,
		    kIOUCScalarIScalarO, 0, 1);
  SetExternalMethod(kAppleBt8xxCmdSetSource, this, _extCmdSetSource,
		    kIOUCScalarIScalarO, 1, 0);
  
  SetExternalMethod(kAppleBt8xxCmdSetTuner, this, _extCmdNone,
		    kIOUCScalarIScalarO, 0, 0);
  SetExternalMethod(kAppleBt8xxCmdSetTuner, this, _extCmdSetTuner,
		    kIOUCScalarIScalarO, 1, 0);
  
  SetExternalMethod(kAppleBt8xxCmdSetVideoControl, this,
		    _extCmdSetVideoControl,
		    kIOUCScalarIScalarO, 2, 0);
  
  SetExternalMethod(kAppleBt8xxCmdReadReg, this, _extCmdReadReg,
		    kIOUCScalarIScalarO, 1, 1);
  SetExternalMethod(kAppleBt8xxCmdWriteReg, this, _extCmdWriteReg,
		    kIOUCScalarIScalarO, 2, 0);
  
  owner = (AppleBt8xx *)provider;
  
  return true;
}

IOReturn AppleBt8xxUserClient::clientClose(void)
{
  owner->close();
  
  detach(owner);
  
  return kIOReturnSuccess;
}

IOService *AppleBt8xxUserClient::getService(void)
{
  return owner;
}

IOExternalMethod *AppleBt8xxUserClient::getTargetAndMethodForIndex(IOService **targetP, UInt32 index)
{
  IOExternalMethod *method = NULL;
  
  if (index < kAppleBt8xxCmdCount) {
    method = externals + index;
    *targetP = (IOService *)method->object;
  }
  
  return method;
}
