/*
 * Copyright (c) 1998-2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 *
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 */
/*
 * Copyright (c) 2000-2001 Apple Computer, Inc.  All rights reserved.
 *
 *  DRI: Josh de Cesare
 *
 */

#include <IOKit/IOService.h>
#include <IOKit/IOUserClient.h>
#include <IOKit/pci/IOPCIDevice.h>

#include "AppleBt8xxCmds.h"

// Bt8xx chip types
enum {
    kBt8xx = 0,
    kBt848,
    kBt848A,
    kBt849A,
    kBt878,
    kBt878a,
    kBt879
};

class AppleBt8xxUserClient;

class AppleBt8xx : public IOService
{
  OSDeclareDefaultStructors(AppleBt8xx);
  friend class AppleBt8xxUserClient;
  
public:
  UInt32	   chipType;
  const OSSymbol   *uniqueSymbol;
  IOPCIDevice      *videoNub;
  IOPCIDevice      *audioNub;
  IOMemoryMap      *videoMemoryMap;
  IOMemoryMap      *audioMemoryMap;
  UInt8            *videoBaseAddress;
  UInt8            *audioBaseAddress;
  
  static  void metaInit(void);
  virtual bool init(OSDictionary *dictionary = 0);
  virtual void free(void);
  
  virtual bool attach(IOService *provider);
  virtual void detach(IOService *provider);
  
  virtual IOService *probe(IOService *provider, SInt32 *score);
  
  virtual bool start(IOService *provider);
  virtual void stop(IOService *provider);
  
  virtual IOReturn newUserClient(task_t owningTask, void *security_id,
				 UInt32 type, IOUserClient **handler);
  virtual void close(void);
  
  virtual IOReturn setProperties(OSObject *properties);
  
private:
  OSArray              *_sources;
  
  IOUserClient         *_userClient;
  
  eBtMode              _currentMode;
  eBtFormat            _currentFormat;
  
  SInt32               _currentSource;
  bool                 _currentSourceHasVideo;
  bool                 _currentSourceHasAudio;
  OSDictionary         *_currentSourceProperties;
  
  UInt32               _audioMask;
  
  UInt8                *_videoCmdPage;
  UInt8                *_videoEvenPage;
  UInt8                *_videoOddPage;
  IOPhysicalAddress    _videoCmdPagePhys;
  IOPhysicalAddress    _videoEvenPagePhys;
  IOPhysicalAddress    _videoOddPagePhys;
  
  AppleBt8xxScreenInfo _screenInfo;
  
  UInt32               _left, _top, _width, _heigth;
  
  void		_identifyChipType(IOPCIDevice *pciNub);

  IOReturn      _processCmd(AppleBt8xxCmdPtr cmd);
  
  IOReturn     _reset(void);
  IOReturn     _setup(void);
  IOReturn     _start(void);
  IOReturn     _stop(void);
  
  OSDictionary *_getSourceProperties(UInt32 source);
  IOReturn     _setSource(SInt32 source);
  
  void         _selectFormat(eBtFormat format);
  
  IOReturn     _setTuner(UInt32 freq);
  
  IOReturn     _setBrightness(SInt8 brightness);
  IOReturn     _setContrast(UInt16 contrast);
  IOReturn     _setSaturationU(UInt16 saturationU);
  IOReturn     _setSaturationV(UInt16 saturationV);
  IOReturn     _setHue(SInt8 hue);
  
  IOReturn     _sendIICWrite(UInt8 addr, UInt8 data1, UInt8 data2, bool three);
  IOReturn     _sendIICRead(UInt8 addr, UInt8 subaddr, UInt8 *data, UInt8 length);
  
  // DMA programing methods
  void         _setupVideoDMA(void);
  
  // Register accessor methods
  UInt8        _readVideoReg8(UInt32 reg);
  UInt32       _readVideoReg32(UInt32 reg);
  void         _writeVideoReg8(UInt32 reg, UInt8 data);
  void         _writeVideoReg32(UInt32 reg, UInt32 data);
  UInt8        _readAudioReg8(UInt32 reg);
  UInt32       _readAudioReg32(UInt32 reg);
  void         _writeAudioReg8(UInt32 reg, UInt8 data);
  void         _writeAudioReg32(UInt32 reg, UInt32 data);
};

class AppleBt8xxUserClient : public IOUserClient
{
  OSDeclareDefaultStructors(AppleBt8xxUserClient);
  
private:
  AppleBt8xx       *owner;
  IOExternalMethod externals[kAppleBt8xxCmdCount];
  
public:
  static AppleBt8xxUserClient *withTask(task_t owningTask);
  
  virtual bool start(IOService *provider);
  
  virtual IOReturn clientClose(void);
  virtual IOService *getService(void);
  virtual IOExternalMethod *getTargetAndMethodForIndex(IOService **targetP,
						       UInt32 index);
  
private:
  IOReturn _extCmdNone(void);
  IOReturn _extCmdSetup(AppleBt8xxCmdSetupInfoPtr info);
  IOReturn _extCmdStart(void);
  IOReturn _extCmdStop(void);
  IOReturn _extCmdGetSource(SInt32 *source);
  IOReturn _extCmdSetSource(SInt32 source);
  IOReturn _extCmdSetTuner(UInt32 frequency);
  IOReturn _extCmdSetVideoControl(eAppleBt8xxControl control, UInt32 value);
  IOReturn _extCmdReadReg(UInt32 reg, UInt32 *data);
  IOReturn _extCmdWriteReg(UInt32 reg, UInt32 data);
};
