/*
 * Copyright (c) 2000 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 *
 * The contents of this file constitute Original Code as defined in and
 * are subject to the Apple Public Source License Version 1.1 (the
 * "License").  You may not use this file except in compliance with the
 * License.  Please obtain a copy of the License at
 * http://www.apple.com/publicsource and read it before using this file.
 *
 * This Original Code and all software distributed under the License are
 * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
 * License for the specific language governing rights and limitations
 * under the License.
 *
 * @APPLE_LICENSE_HEADER_END@
 *
 * TellyController.m
 * Wilfredo Sanchez | wsanchez@apple.com
 */

#import "TellyController.h"

#include "AppleBt8xxCmds.h"

@implementation TellyController

/**
 * Inits
 **/

- (void) _initVideoCard
{
    io_iterator_t aCardIterator = NULL;
    kern_return_t anError       = KERN_SUCCESS;
    mach_port_t   aMasterPort   = NULL;
    
    myVideoCard = NULL;
    
    if (anError == KERN_SUCCESS)
	anError = IOMasterPort(bootstrap_port, &aMasterPort);
    
    if (anError == KERN_SUCCESS)
      {
	IOServiceGetMatchingServices(aMasterPort,
					IOServiceNameMatching("AppleBt8xx"),
					&aCardIterator);
    
	// FIXME: Deallocate the master port
    
	if (anError == KERN_SUCCESS)
	    myVideoCard = IOIteratorNext(aCardIterator);
    
	if (aCardIterator)
	    IOObjectRelease(aCardIterator);
      }

    if (myVideoCard)
        anError = IOServiceOpen(myVideoCard, mach_task_self(), 0, &myVideoConnect);
    
    if (! myVideoCard || anError != KERN_SUCCESS)
      {
	NSRunCriticalAlertPanel(@"Warning",
				@"No suitable video card was found.",
				@"Bummer", nil, nil);
	myInputSourceInfo = nil;
      }
    else
      {
	myInputSourceInfo =
	    (NSArray*)IORegistryEntryCreateCFProperty(myVideoCard,
						    (CFStringRef)@"AppleBt8xxSources",
						    NULL, 0);
      }
}

- (void) _initChannelMap
{
    NSString* aChannelFile =
	[[NSBundle mainBundle] pathForResource: @"ChannelMap" ofType: @"plist"];
    NSString* myChannelFileText =
	[NSString stringWithContentsOfFile: aChannelFile];

    if (! aChannelFile     ) NSLog(@"Channel map file missing.");
    if (! myChannelFileText) NSLog(@"Channel map file unreadable.");

    if ((myChannelMap = [myChannelFileText propertyList]))
      {
	if ([myChannelMap isKindOfClass:[NSDictionary class]])
	    [myChannelMap retain];
	else
	  {
	    NSLog(@"Channel map is not a dictionary.");
	    myChannelMap = nil;
	  }
      }
    else
	NSLog(@"Can't parse channel map file.");

    if (! myChannelMap)
      {
	NSRunCriticalAlertPanel(@"Error",
				@"Channel map not found (installation error?).",
				@"Quit", nil, nil);
	[NSApp terminate: self];
      }
}

- (id) init
{
    if ((self = [super init]))
      {
	[self _initVideoCard ];
	[self _initChannelMap];
      }
    return self;
}

- (void) dealloc
{
    if (myChannelMap     ) [myChannelMap      release];
    if (myInputSourceInfo) [myInputSourceInfo release];

    if (myVideoConnect) IOServiceClose  (myVideoConnect);
    if (myVideoCard   ) IOObjectRelease (myVideoCard   );

    [super dealloc];
}

- (void) _initInputSourcePopUp
{
    /* Initialize input source pop-up */
    if (myInputSourceInfo)
      {
	int aSourceCount = [myInputSourceInfo count];
	int aSourceIndex;

	[tellyInputSourcePopUp removeAllItems];

	for (aSourceIndex = 0; aSourceIndex < aSourceCount; aSourceIndex++)
	  {
	    NSMutableDictionary* aSource        = [myInputSourceInfo objectAtIndex: aSourceIndex];
            NSArray*             aSubFormatList = [aSource objectForKey: @"SubFormats"];

            if (aSubFormatList)
              {
		NSEnumerator* aSubFormatEnumerator = [aSubFormatList objectEnumerator];
		NSString*     aSubFormat;

		while (aSubFormat = [aSubFormatEnumerator nextObject])
		  {
		    NSMutableDictionary* aSourceCopy = [[myInputSourceInfo objectAtIndex: aSourceIndex] mutableCopy];

		    /* Replace the array of sub-formats with a singe subformat */
		    [aSourceCopy removeObjectForKey: @"SubFormats"];
		    [aSourceCopy setObject: aSubFormat forKey: @"SubFormat"];

		    /* Keep track of the index of the original format */
                    [aSourceCopy setObject: [NSNumber numberWithInt: aSourceIndex] forKey: @"SourceIndex"];

		    /* Add to pop-up list */
                    [tellyInputSourcePopUp addItemWithTitle: aSubFormat];
                    [[tellyInputSourcePopUp lastItem] setRepresentedObject: aSourceCopy];

		    [aSourceCopy release];
		  }
              }
	    else
		NSLog(@"No names in input source %@", aSource);
	  }
      }
}

- (void) _initChannelPopUp
{
    NSDictionary* aSource = [[tellyInputSourcePopUp selectedItem] representedObject];

    if (aSource)
      {
	NSString*     aFormat;
	NSDictionary* aFormatMap;
        NSString*     aSubFormat; 

	NSDictionary* aFrequencyMap = nil;

	NSEnumerator* aFrequencyEnumerator;
	NSString*     aChannel;

	[tellyChannelPopUp removeAllItems         ];
	[tellyChannelPopUp addItemWithTitle : @"-"];
	[tellyChannelPopUp setEnabled       : NO  ];

	if (!(aFormat       = [aSource      objectForKey: @"Format"   ])) { NSLog(@"No format for source %@"          , aSource   ); return; }
	if (!(aSubFormat    = [aSource      objectForKey: @"SubFormat"])) { NSLog(@"No sub-format for source %@"      , aSource   ); return; }
	if (!(aFormatMap    = [myChannelMap objectForKey: aFormat     ])) { NSLog(@"No channel map for format %@."    , aFormat   ); return; }
	if (!(aFrequencyMap = [aFormatMap   objectForKey: aSubFormat  ])) { NSLog(@"Frequency map for sub-format %@"  , aSubFormat); return; }

	[tellyChannelPopUp removeAllItems];

	aFrequencyEnumerator =
	    [[[aFrequencyMap allKeys] sortedArrayUsingSelector: @selector(compare:)] objectEnumerator];

	while (aChannel = [aFrequencyEnumerator nextObject])
	  {
	    NSString* aFrequency = [aFrequencyMap objectForKey: aChannel];

	    [tellyChannelPopUp addItemWithTitle: [NSString stringWithFormat: @"Channel %@", aChannel]];

	    [[tellyChannelPopUp lastItem] setRepresentedObject: aFrequency];
	  }

	[tellyChannelPopUp setEnabled: YES];

	[self frobChannel: self];
      }
    else
	NSLog(@"Invalid input source %d", [self inputSource]);
}

- (void) awakeFromNib
{
    if (! controlsDrawer        ||
	! tellyView             ||
	! tellyWindow           ||
	! tellyPowerButton      ||
	! tellyInputSourcePopUp ||
	! tellyChannelFormCell  ||
	! tellyHueSlider        ||
	! tellySaturationSlider ||
	! tellyBrightnessSlider ||
	! tellyContrastSlider   )
	{
	    NSRunCriticalAlertPanel(@"Error",
				    @"Telly window failed to initialize.",
				    @"Quit", nil, nil);
	    [NSApp terminate: self];
	}

    /* Init UI elements */
    [self _initInputSourcePopUp];

    /* Init iVars based on default UI state */
    [self frobPower       : self];
    [self frobInputSource : self]; // Causes frobChannel
    [self frobHue         : self];
    [self frobSaturation  : self];
    [self frobContrast    : self];
    [self frobBrightness  : self];
}

/**
 * Actions
 **/

/* Private actions */

- (void) _updateWindowPosition
{
    // FIXME: Figure out position of tellyView

    NSPoint aViewOriginInWindow    = [tellyView convertPoint: NSZeroPoint
						      toView: [[tellyView window] contentView]];
    NSPoint aWindowOriginInDisplay = [[tellyView window] frame].origin;
    NSPoint aViewOriginInDisplay   = NSMakePoint(aWindowOriginInDisplay.x + aViewOriginInWindow.x,
						 aWindowOriginInDisplay.y + aViewOriginInWindow.y);

    NSLog(@"Telly view origin in window    : (%f, %f)", aViewOriginInWindow.x    , aViewOriginInWindow.y   );
    NSLog(@"Telly window origin in display : (%f, %f)", aWindowOriginInDisplay.x , aWindowOriginInDisplay.y) ;
    NSLog(@"Telly view origin in display   : (%f, %f)", aViewOriginInDisplay.x   , aViewOriginInDisplay.y  ) ;

    // FANCY FIX: Figure out clipping regions, too

    // Notify the driver of screen position here
}

/**
 * Accessors
 **/

- (BOOL     ) power       { return [tellyPowerButton       state              ]; }
- (int      ) inputSource { return [tellyInputSourcePopUp  indexOfSelectedItem]; }
- (float    ) frequency   { return [tellyFrequencyFormCell floatValue         ]; }
- (NSString*) channel     { return @""                                         ; }
- (int      ) hue         { return [tellyHueSlider         intValue           ]; }
- (int      ) saturation  { return [tellySaturationSlider  intValue           ]; }
- (int      ) contrast    { return [tellyContrastSlider    intValue           ]; }
- (int      ) brightness  { return [tellyBrightnessSlider  intValue           ]; }

- (void) setPower: (BOOL) aValue
{
    kern_return_t anError;

    if (aValue)
      {
	[self _updateWindowPosition];

	anError = AppleBt8xxCmdStart(myVideoConnect);
      }
    else
	anError = AppleBt8xxCmdStop(myVideoConnect);

    if (anError == KERN_SUCCESS)
	[tellyPowerButton setState: aValue];
    else
	NSLog(@"Error setting power");
}

- (void) setInputSource: (int) aValue
{
    NSNumber* aSourceNumber = 
	[[[tellyInputSourcePopUp itemAtIndex: aValue] representedObject] objectForKey: @"SourceIndex"];

    if (aSourceNumber)
      {
        SInt32 aSource = (SInt32)[aSourceNumber intValue];

	if (AppleBt8xxCmdSetSource(myVideoConnect, aSource) == KERN_SUCCESS)
	  {
	    [tellyInputSourcePopUp selectItemAtIndex: aValue];

	    [self _initChannelPopUp];
	  }
	else
	    NSLog(@"Error setting input source");
      }
    else
	NSLog(@"No input source %d.", aValue);
}

- (void) setFrequency: (float) aValue
{
    if (AppleBt8xxCmdSetTuner(myVideoConnect, (UInt32)(aValue * 100)) == KERN_SUCCESS)
      {
	[tellyFrequencyFormCell setFloatValue: aValue];

	// FIXME: Also set pop-up
      }
    else
	NSLog(@"Error setting frequency");
}

- (void) setChannel: (NSString*) aValue
{
    NSLog(@"WARNING: setChannel: not implemented");

    [tellyChannelFormCell setObjectValue: @""];
}

- (void) setHue: (int) aValue
{
    [tellyHueSlider setIntValue: aValue];

    AppleBt8xxCmdSetVideoControl(myVideoConnect, kAppleBt8xxControlHue, aValue);
}

- (void) setSaturation: (int) aValue
{
    [tellySaturationSlider setIntValue: aValue];

    AppleBt8xxCmdSetVideoControl(myVideoConnect, kAppleBt8xxControlSaturationU, aValue);
    AppleBt8xxCmdSetVideoControl(myVideoConnect, kAppleBt8xxControlSaturationV, aValue);
}

- (void) setContrast: (int) aValue
{
    [tellyContrastSlider setIntValue: aValue];

    AppleBt8xxCmdSetVideoControl(myVideoConnect, kAppleBt8xxControlContrast, aValue);
}

- (void) setBrightness: (int) aValue
{
    [tellyBrightnessSlider setIntValue: aValue];

    AppleBt8xxCmdSetVideoControl(myVideoConnect, kAppleBt8xxControlBrightness, aValue);
}

/**
 * Telly window delegate methods
 **/

- (void) windowWillClose       : (NSNotification*) aNotification { myPowerBeforeMove = [self power]; [self setPower: NO]; }
- (void) windowWillMiniaturize : (NSNotification*) aNotification { myPowerBeforeMove = [self power]; [self setPower: NO]; }
- (void) windowWillMove        : (NSNotification*) aNotification { myPowerBeforeMove = [self power]; [self setPower: NO]; }

- (void) windowDidDeminiaturize : (NSNotification*) aNotification { [self setPower: myPowerBeforeMove]; }
- (void) windowDidMove          : (NSNotification*) aNotification { [self setPower: myPowerBeforeMove]; }

/**
 * IB controls
 **/

/* Knob frobbers */
- (IBAction) frobPower       : (id) aSender { [self setPower       : [self power      ]]; }
- (IBAction) frobInputSource : (id) aSender { [self setInputSource : [self inputSource]]; }
- (IBAction) frobHue         : (id) aSender { [self setHue         : [self hue        ]]; }
- (IBAction) frobSaturation  : (id) aSender { [self setSaturation  : [self saturation ]]; }
- (IBAction) frobContrast    : (id) aSender { [self setContrast    : [self contrast   ]]; }
- (IBAction) frobBrightness  : (id) aSender { [self setBrightness  : [self brightness ]]; }

- (IBAction) frobChannel: (id) aSender
{
    if (aSender == tellyChannelPopUp || aSender == self)
	[self setFrequency: [[[tellyChannelPopUp selectedItem] representedObject] floatValue]];
    else if (aSender == tellyFrequencyFormCell)
	[self setFrequency: [tellyFrequencyFormCell floatValue]];
    else if (aSender == tellyChannelFormCell)
	[self setChannel: [tellyChannelFormCell stringValue]];
    else
      {
	NSLog(@"Unknown sender %@ requested channel change.", aSender);
	[self setFrequency: [tellyFrequencyFormCell floatValue]];
      }
}

/* Drawer actions */
- (void) toggleDrawer : (id) aSender
{
    switch ([controlsDrawer state])
      {
	case NSDrawerClosedState:
	case NSDrawerClosingState:
	    [controlsDrawer open];
	    break;
	case NSDrawerOpenState:
	case NSDrawerOpeningState:
	    [controlsDrawer close];
	    break;
	default:
	    NSLog(@"Wuh? Unknown drawer state...");
	    [controlsDrawer open];
	    break;
      }
}

@end
